import logging
import json

from allauth.account.models import (
    EmailAddress,
)
from django.contrib.auth import get_user_model
from django.core import mail
from django.urls import reverse
from django.test.utils import override_settings

from rest_framework import status

from django_sso_app.backend.tests.factories import UserTestCase
from django_sso_app.core import app_settings

User = get_user_model()
logger = logging.getLogger('django_sso_app.core.tests')


class TestLogin(UserTestCase):

    def test_can_login_with_email(self):
        response = self.client.post(
            reverse('rest_login'),
            data = json.dumps(self.valid_email_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_can_login_with_username(self):
        if 'username' in app_settings.USER_FIELDS:
            response = self.client.post(
                reverse('rest_login'),
                data = json.dumps(self.valid_username_login),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_login_returns_jwt_cookie(self):
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_email_login),
            content_type='application/json'
        )

        cookies = response.client.cookies.items()

        has_jwt_cookie = False
        for (key, val) in cookies:
            if key == "jwt":
                has_jwt_cookie = True
                break

        self.assertTrue(has_jwt_cookie)

    def test_migrated_can_login_without_updating_rev(self):
        self.client.logout()

        migrated_user = User.objects.create(
            username=self.migrated_user_username,
            email=self.migrated_user_email,
            password=self.migrated_user_hashed_password
        )
        migrated_user_email = EmailAddress.objects.create(user=migrated_user,
                                                          email=self.migrated_user_email,
                                                          primary=True,
                                                          verified=True)
        migrated_profile = migrated_user.sso_app_profile

        logger.info('User created "{}"'.format(migrated_profile))

        previus_rev = migrated_profile.sso_rev

        logger.info('Calling Login with credentials "{}"'.format(self.valid_migrated_user_login))
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_migrated_user_login),
            content_type='application/json'
        )

        logger.info('resp!!!! "{}" "{}"'.format(response, response.data))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        migrated_profile.refresh_from_db()
        second_rev = migrated_profile.sso_rev
        self.assertEqual(previus_rev, second_rev)

    def test_django_superuser_can_not_login_from_api_views(self):
        response = self.client.post(
            reverse('rest_login'),
            data=self.valid_admin_login
        )

        cookies = response.client.cookies.items()

        jwt_cookie = None
        has_jwt_cookie = False
        for (key, val) in cookies:
            if key == "jwt":
                has_jwt_cookie = True
                jwt_cookie = val
                break

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertFalse(has_jwt_cookie, 'superuser receives jwt')
        self.assertEqual(jwt_cookie, None, 'jwt is still valid')

    def test_django_staff_user_can_not_login_from_api_views(self):
        response = self.client.post(
            reverse('rest_login'),
            data=self.valid_staff_login
        )

        cookies = response.client.cookies.items()

        jwt_cookie = None
        has_jwt_cookie = False
        for (key, val) in cookies:
            if key == "jwt":
                has_jwt_cookie = True
                jwt_cookie = val
                break

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertFalse(has_jwt_cookie, 'staff receives jwt')
        self.assertEqual(jwt_cookie, None, 'jwt is still valid')


class TestLogout(UserTestCase):

    @override_settings(DJANGO_SSO_APP_LOGOUT_DELETES_ALL_PROFILE_DEVICES=True)
    def test_api_logout_deletes_jwt_and_removes_all_profile_devices(self):
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_username_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        user_devices_pre_logout = self.user.sso_app_profile.devices.all()

        response = self.client.post(
            reverse('rest_logout'),
            content_type='application/json'
        )

        cookies = response.client.cookies.items()
        print('COOKIE', cookies)
        jwt_cookie = None
        has_valid_jwt_cookie = False
        for (key, val) in cookies:
            if key == "jwt":
                jwt_cookie = val
                if getattr(jwt_cookie, '_value') != 'None':
                    has_valid_jwt_cookie = True
                break

        self.assertFalse(has_valid_jwt_cookie, 'api logout keeps jwt cookie')

        user_devices_post_logout = self.user.sso_app_profile.devices.all()

        self.assertEqual(user_devices_post_logout.count(), 0, 'api logout is not deleting all profile devices')

    @override_settings(DJANGO_SSO_APP_LOGOUT_DELETES_ALL_PROFILE_DEVICES=False)
    def test_api_logout_deletes_jwt_and_removes_caller_device(self):
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_username_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        user_devices_pre_logout = self.user.sso_app_profile.devices.all()

        response = self.client.post(
            reverse('rest_logout'),
            content_type='application/json'
        )

        cookies = response.client.cookies.items()
        print('COOKIE', cookies)
        jwt_cookie = None
        has_valid_jwt_cookie = False
        for (key, val) in cookies:
            if key == "jwt":
                jwt_cookie = val
                if getattr(jwt_cookie, '_value') != 'None':
                    has_valid_jwt_cookie = True
                break

        self.assertFalse(has_valid_jwt_cookie, 'api logout keeps jwt cookie')

        user_devices_post_logout = self.user.sso_app_profile.devices.all()

        self.assertEqual(user_devices_post_logout.count(), user_devices_pre_logout.count(),
                         'api logout is deleting all user devices')


class TestSignup(UserTestCase):

    def test_can_register(self):
        response = self.client.post(
            reverse('rest_signup'),
            data=json.dumps(self.valid_new_profile),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        new_user = User.objects.filter(email=self.valid_new_profile['email']).first()

        self.assertEqual(new_user.email, self.valid_new_profile['email'])

        email_address = EmailAddress.objects.filter(user=new_user, email=self.valid_new_profile['email']).first()

        self.assertNotEqual(email_address, None)

    def test_cannot_register_on_invalid_registration_data(self):
        response = self.client.post(
            reverse('rest_signup'),
            data = json.dumps(self.invalid_new_profile),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_registration_sends_confirmation_email(self):
        response = self.client.post(
            reverse('rest_signup'),
            data = json.dumps(self.valid_new_profile),
            content_type='application/json'
        )

        self.assertEqual(len(mail.outbox), 1)



import json
import jwt

from django.contrib.auth import get_user_model
from django.test import Client
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.core import mail

from allauth.account.adapter import get_adapter
from rest_framework import status

from django_sso_app.backend.tests.factories import UserTestCase

from .... import app_settings

User = get_user_model()


class TestUserCreate(UserTestCase):

    def test_staff_user_can_create_user_with_profile_with_unverified_email(self):
        print('HEADERS', self.staff_user_valid_token_headers, type(self.staff_user_valid_token_headers))
        response = self.client.post(
            reverse('django_sso_app_user:rest-list'),
            data=json.dumps(self.valid_new_external_user),
            **self.staff_user_valid_token_headers
        )

        new_user_sso_id = response.data.get('sso_id')

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        response2 = self.client.get(
            reverse('django_sso_app_user:rest-detail', args=(new_user_sso_id,)),
            **self.staff_user_valid_token_headers
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        self.assertEqual(response2.data['email'], self.valid_new_external_user['email'])

        self.assertEqual(response2.data['email_verified'], False)

    def test_staff_user_can_create_user_with_profile_and_validated_email(self):
        print('HEADERS', self.staff_user_valid_token_headers, type(self.staff_user_valid_token_headers))

        response = self.client.post(
            reverse('django_sso_app_user:rest-list') + '?skip_confirmation=true',
            data=json.dumps(self.valid_new_external_user),
            **self.staff_user_valid_token_headers
        )

        new_user_sso_id = response.data.get('sso_id')

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        response2 = self.client.get(
            reverse('django_sso_app_user:rest-detail', args=(new_user_sso_id,)),
            **self.staff_user_valid_token_headers
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)


        self.assertEqual(response2.data['email'], self.valid_new_external_user['email'])

        self.assertEqual(response2.data['email_verified'], True)

    def test_non_staff_user_can_create_profile_without_validated_mail(self):
        response0 = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )

        response = self.client.post(
            reverse('django_sso_app_user:rest-list') + '?skip_confirmation=true',
            data=json.dumps(self.valid_new_external_user),
            content_type='application/json'
        )

        print('RESP', response.json())

        new_user_sso_id = response.data.get('sso_id')

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        response2 = self.client.get(
            reverse('django_sso_app_user:rest-detail', args=(new_user_sso_id,)),
            **self.staff_user_valid_token_headers
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        self.assertEqual(response2.data['email'], self.valid_new_external_user['email'])

        self.assertEqual(response2.data['email_verified'], False)

        new_user = User.objects.get(sso_app_profile__sso_id=new_user_sso_id)
        print('USER SUBSCRIPT', new_user.sso_app_profile.subscriptions.all(), self.service2.__dict__)

        self.assertEqual(new_user.sso_app_profile.subscriptions.first().service, self.service2 ,
                         'no subscription created')

        profile_object = self.valid_new_external_user['profile']
        for f in app_settings.PROFILE_FIELDS:
            model_val = getattr(new_user.sso_app_profile, f, None)
            object_val = profile_object.get(f, None)

            if f == 'birthdate':
                model_val = str(model_val)
            elif f == 'country':
                object_val = object_val.upper()
                model_val = str(model_val).upper()

            self.assertEqual(model_val, object_val)


class TestUserRetrieve(UserTestCase):

    def test_user_can_retrieve_user(self):
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = self.client.get(
            reverse('django_sso_app_user:rest-detail', args=(self.profile.sso_id,)),
            content_type='application/json'
        )
        print('RESP', response2.data)
        self.assertEqual(response2.status_code, status.HTTP_200_OK)

    def test_staff_user_can_retrieve_user_with_token(self):

        response2 = self.client.get(
            reverse('django_sso_app_user:rest-detail', args=(self.profile.sso_id,)),
            **self.staff_user_valid_token_headers
        )
        print('RESP', response2.data)
        self.assertEqual(response2.status_code, status.HTTP_200_OK)


class TestUserUpdate(UserTestCase):

    def test_user_email_update_deletes_all_devices_and_waits_for_confirmation(self):
        print('\n logging in with user', self.user, 'and fingerprint', self.valid_login.get('fingerprint'))

        first_rev = self.profile.sso_rev
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        print('\n sending patch from user', self.user, 'with new email', self.valid_email_update)
        response2 = self.client.patch(
            reverse('django_sso_app_user:rest-detail', args=[self.profile.sso_id]),
            data=json.dumps(self.valid_email_update),
            content_type='application/json'
        )
        cookies2 = response2.cookies

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        devices = self.profile.devices.all()
        self.assertEqual(devices.count(), 0, 'devices not deleted')

        self.assertEqual(len(mail.outbox), 1, 'no email sent')

        print('Emaiol', mail.outbox)
        cofirmation_message = mail.outbox[0]
        recipients = cofirmation_message.recipients()
        self.assertIn(self.valid_email_update['email'], recipients, 'user not in recipients')

        client = Client()
        response3 = client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_email_login),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, 200, 'unable to login with old email')

        client = Client()
        response4 = client.post(
            reverse('rest_login'),
            data=json.dumps(self.invalid_unconfirmed_email_login),
            content_type='application/json'
        )
        self.assertFalse(self.user.emailaddress_set.get(email=self.invalid_unconfirmed_email_login['login']).verified)
        self.assertEqual(response4.status_code, 400, 'login with unconfirmed email is still possible')

        unconfirmed_email = self.user.emailaddress_set.get(email=self.invalid_unconfirmed_email_login['login'])
        self.assertEqual(unconfirmed_email.verified, False)

        adapter = get_adapter()
        adapter.confirm_email(None, unconfirmed_email)

        client = Client()
        response5 = client.post(
            reverse('rest_login'),
            data=json.dumps(self.invalid_unconfirmed_email_login),
            content_type='application/json'
        )
        self.assertEqual(response5.status_code, 200, 'login with confirmed email is not possible')

        self.assertEqual(self.user.emailaddress_set.count(), 1)

        client = Client()
        response6 = client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_email_login),
            content_type='application/json'
        )

        self.assertEqual(response6.status_code, 400, 'login old email is still possible')

    def test_user_password_update_refreshes_and_returns_current_device_jwt_and_deletes_all_devices_but_the_caller(self):
        first_rev = self.profile.sso_rev
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies
        self.assertEqual(response.status_code, 200, 'unable to login with valid credentials')

        response2 = self.client.patch(
            reverse('django_sso_app_user:rest-detail', args=[self.profile.sso_id]),
            data=json.dumps(self.valid_password_update),
            content_type='application/json'
        )
        cookies2 = response2.cookies

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertTrue(len(cookies2.keys()) > 0, 'no jwt cookie set')

        devices = self.profile.devices.all()

        self.assertEqual(devices.count(), 1, 'other devices not deleted')

        self.assertEqual(devices.first().fingerprint, self.valid_login['fingerprint'], 'device not updated')

        jwt_cookie = cookies.get('jwt').value
        jwt_cookie2 = cookies2.get('jwt').value

        # print('\nRISPOSTA, user rev', self.profile.sso_rev, '\n\nTOKEN', jwt_cookie)

        unverified_payload = jwt.decode(jwt_cookie, None, False)
        unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

        self.assertEqual(unverified_payload2.get('sso_rev'), unverified_payload.get('sso_rev') + 1,
                         'rev not incremented in cookie')

        response3 = Client().post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login_after_password_change),
            content_type='application/json'
        )

        cookies3 = response3.cookies
        #print('\n\n\n RISPOSTS!', response3.data, cookies3)

        self.assertEqual(response3.status_code, 200, 'unable to login with new password')
        jwt_cookie3 = cookies3.get('jwt').value
        unverified_payload3 = jwt.decode(jwt_cookie3, None, False)

        print('\n\n\n Unv pay', unverified_payload3)

        self.assertEqual(unverified_payload2.get('sso_rev'), unverified_payload3.get('sso_rev'), 'rev changed after login')

    def test_user_password_update_refreshes_profile_even_if_called_with_authorization_headers(self):
        """
        User password update
        """

        first_rev = self.profile.sso_rev
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        print('RESPONSE1', response)

        cookies = response.cookies

        self.assertEqual(response.status_code, 200, 'unable to login with valid credentials')

        received_jwt = cookies.get('jwt').value

        self.assertEqual(response.data['token'], received_jwt, 'jwt differs between cookie and response')

        response2 = Client().patch(
            '/api/v1/auth/users/{0}/'.format(self.profile.sso_id),
            data=json.dumps(self.valid_password_update),
            content_type='application/json',
            HTTP_AUTHORIZATION='Bearer {}'.format(received_jwt)
        )

        print('RESPONSE2', self.valid_password_update, response2.data)

        self.assertEqual(response2.status_code, 200, 'cannot patch with jwt inside headers')
        cookies2 = response2.cookies

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertTrue(len(cookies2.keys()) > 0, 'no jwt cookie set')

        devices = self.profile.devices.all()

        self.assertEqual(devices.count(), 1, 'other devices not deleted')

        self.assertEqual(devices.first().fingerprint, self.valid_login['fingerprint'], 'device not updated')

        jwt_cookie = cookies.get('jwt').value
        jwt_cookie2 = cookies2.get('jwt').value

        # print('\nRISPOSTA, user rev', self.profile.sso_rev, '\n\nTOKEN', jwt_cookie)

        unverified_payload = jwt.decode(jwt_cookie, None, False)
        unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

        self.assertEqual(unverified_payload2.get('sso_rev'), unverified_payload.get('sso_rev') + 1,
                         'rev not incremented in cookie')

        print('LOGGING IN')
        client3 = Client()
        response3 = client3.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login_after_password_change),
            content_type='application/json'
        )

        cookies3 = response3.cookies

        self.assertEqual(response3.status_code, 200, 'unable to login with new password')

        jwt_cookie3 = cookies3.get('jwt').value
        unverified_payload3 = jwt.decode(jwt_cookie3, None, False)

        self.assertEqual(unverified_payload2.get('sso_rev'), unverified_payload3.get('sso_rev'), 'rev changed after login')

    def test_user_email_update_by_staff_disables_user_login_user_must_confirm_email(self):
        """
        User update by staff refreshes disables login and deletes all user devices
        """
        first_rev = self.profile.sso_rev

        response2 = self.client.patch(
            '/api/v1/auth/users/{0}/'.format(self.profile.sso_id),
            data=json.dumps(self.valid_email_update),
            **self.staff_user_valid_token_headers
        )
        cookies2 = response2.cookies

        self.assertTrue('jwt' not in cookies2.keys(), 'jwt cookie set')

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        devices = self.profile.devices.all()

        self.assertEqual(devices.count(), 0, 'user devices not deleted')

        response3 = Client().post(
            reverse('rest_login'),
            data=json.dumps(self.invalid_unconfirmed_email_login),
            content_type='application/json'
        )
        cookies3 = response3.cookies
        print('RESP!!!!', response3.data)

        self.assertEqual(response3.status_code, status.HTTP_400_BAD_REQUEST,
                         'user can login without confirmating email change')

        self.assertEqual(response3.data, [_(
            'We have sent an e-mail to you for verification. Follow the link provided to finalize the signup process. Please contact us if you do not receive it within a few minutes.')])

        response4 = Client().post(
            reverse('rest_login'),
            data=json.dumps(self.valid_email_login),
            content_type='application/json'
        )

        self.assertEqual(response4.status_code, status.HTTP_200_OK, 'user can not login with new email')


    def test_user_email_update_by_staff_with_skip_email_confirmation_activates_new_email(self):
        """
        User update by staff refreshes disables login and deletes all user devices
        """
        first_rev = self.profile.sso_rev

        print('\n sending patch from user', self.staff_user, 'to user', self.user, 'with new email',
              self.valid_email_update)
        response2 = self.client.patch(
            '/api/v1/auth/users/{0}/?skip_confirmation=true'.format(self.profile.sso_id),
            data=json.dumps(self.valid_email_update),
            content_type='application/json',
            HTTP_AUTHORIZATION='Token {}'.format(self.staff_user_token.key)
        )
        cookies2 = response2.cookies
        print('Patch response', response2, response2.__dict__)

        self.assertTrue('jwt' not in cookies2.keys(), 'jwt cookie set')

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        devices = self.profile.devices.all()

        self.assertEqual(devices.count(), 0, 'user devices not deleted')

        print('\n Trying to login from user', self.user, 'with credentials', self.valid_login)
        response3 = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies3 = response3.cookies

        print('\n Login from user', self.user, 'returns', response3.data)

        self.assertEqual(response3.status_code, 200, 'user can NOT login without confirmating email change')



import logging

from django.contrib.auth.models import Group as GroupModel
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from ....permissions import is_staff
from ..models import Group
from ...profiles.models import Profile

logger = logging.getLogger('django_sso_app.core.apps.groups.signals')


@receiver(m2m_changed)
def signal_handler_when_user_is_added_or_removed_from_group(action, instance, pk_set, model, **kwargs):

    if model == Group and instance.__class__ == Profile:
        if kwargs.get('raw', False):
            # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
            return

        user = instance.user
        profile = instance

        logger.info('Groups signal')

        is_loaddata = getattr(user, '__dssoa__loaddata', False)
        is_creating = getattr(user, '__dssoa__creating', False)

        must_update_rev = (not is_loaddata) and (not is_creating)

        groups_updated = False

        if action == 'pre_add':
            groups_updated = True
            for pk in pk_set:
                _group = Group.objects.get(id=pk)
                logger.info('Profile "{}" entered group "{}"'.format(profile, _group))

        elif action == 'pre_remove':
            groups_updated = True
            for pk in pk_set:
                _group = Group.objects.get(id=pk)
                logger.info('Profile "{}" exited from group "{}"'.format(profile, _group))

        # updating rev

        if groups_updated and must_update_rev:
            profile.update_rev(True)



from django.contrib.auth import get_user_model

from django_sso_app.backend.tests.factories import UserTestCase
from django_sso_app.core.apps.emails.models import EmailAddress

User = get_user_model()


class TestEmailAddress(UserTestCase):

    def test_user_email_deletes_other_user_emails_when_primary_and_verified(self):
        self.assertEqual(self.user.emailaddress_set.count(), 1)

        new_email = EmailAddress.objects.create(user=self.user, email='email2@example.com')

        self.assertEqual(self.user.emailaddress_set.count(), 2)

        new_email.primary = True
        new_email.verified = True
        new_email.save()

        self.assertEqual(self.user.emailaddress_set.count(), 1, 'other user email not confirmed')



# user groups update updates user rev

from django.test import TestCase
from django.contrib.auth import get_user_model

from ...emails.models import EmailAddress
from ...groups.models import Group
from ...profiles.models import Profile

User = get_user_model()


class UserTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create(username="pippo", email="pippo@disney.com")
        self.user_email = EmailAddress.objects.create(user=self.user,
                                                      email=self.user.email,
                                                      primary=True,
                                                      verified=True)
        self.group = Group.objects.create(name='new_group')
        print('Created group', self.group)


    def test_add_profile_to_group_updates_rev(self):
        user = self.user
        profile = self.user.sso_app_profile

        profile_rev = profile.sso_rev

        profile.groups.add(self.group)

        print('profile groups', profile.groups.all())

        profile.refresh_from_db()

        self.assertEqual(profile.sso_rev, profile_rev + 1)



import os
import datetime
import binascii
from random import seed, randint
import base64

from django.core.management.utils import get_random_secret_key

from .exceptions import ProfileIncompleteException, DectivatedUserException, UnsubscribedUserException, DjangoStaffUsersCanNotLoginException
from .permissions import is_authenticated, is_django_staff
from . import app_settings

# initializing seed
SEED = int(base64.b16encode(bytes(get_random_secret_key(), 'utf-8')), 16)
seed(SEED)


def set_cookie(response, key, value, days_expire=None, secure=True):
    """
    Sets response auth cookie

    :param response:
    :param key:
    :param value:
    :param days_expire:
    :param secure:
    :return:
    """
    if days_expire is None:
        max_age = app_settings.COOKIE_AGE  # django default
    else:
        max_age = days_expire * 24 * 60 * 60

    expires = datetime.datetime.strftime(datetime.datetime.utcnow() + datetime.timedelta(seconds=max_age),
                                         "%a, %d-%b-%Y %H:%M:%S GMT")

    response.set_cookie(key, value, max_age=None, expires=expires, path='/',
                        domain=app_settings.COOKIE_DOMAIN,
                        secure=secure,
                        httponly=app_settings.COOKIE_HTTPONLY)


def invalidate_cookie(response, key):
    """
    Sets response auth cookie to 'Null'

    :param response:
    :param key:
    :return:
    """
    response.set_cookie(key, None, max_age=None, expires='Thu, 01 Jan 1970 00:00:01 GMT', path='/',
                        domain=app_settings.COOKIE_DOMAIN, secure=None, httponly=False)


def set_session_key(request, key, value):
    """
    Sets django sessions request key
    :param request:
    :param key:
    :param value:
    :return:
    """

    #request.session[key] = value

    setattr(request, key, value)


def get_session_key(request, key, default=None):
    """
    Gets django sessions request key
    :param request:
    :param key:
    :param default:
    :return:
    """

    #return request.session.get(key, default)

    return getattr(request, key, default)


def check_user_can_login(user):
    """
    Check user login ability
    :param user:
    :return:
    """

    if is_authenticated(user):
        if is_django_staff(user):
            raise DjangoStaffUsersCanNotLoginException()
        elif not user.is_active:
            raise DectivatedUserException()
        else:
            if user.sso_app_profile.is_unsubscribed:
                raise UnsubscribedUserException()
            else:
                if app_settings.BACKEND_ENABLED and app_settings.BACKEND_SIGNUP_MUST_FILL_PROFILE:
                    if user.sso_app_profile.is_incomplete:
                        raise ProfileIncompleteException('Must fill in profile.')


def is_test_user(user):
    """
    Chack if user is TEST_USER
    :param user:
    :return:
    """

    return user.username == app_settings.TEST_USER_USERNAME or user.email == app_settings.TEST_USER_EMAIL


def get_random_token(length=4):
    """
    Returns random token
    :param request:
    :return:
    """
    start = randint(0, 4095 - length)
    end = start + length

    return binascii.hexlify(os.urandom(2048)).decode("utf-8")[start:end]


def get_random_fingerprint(request):
    """
    Returns random fingerprint
    :param request:
    :return:
    """
    return 'random_{}_{}'.format(id(request), get_random_token(8))



# profile filed update updates rev

from django.contrib.auth import get_user_model

from django_sso_app.backend.tests.factories import UserTestCase

User = get_user_model()


class ProfileTestCase(UserTestCase):

    def test_profile_field_update_updates_rev(self):
        """
        User update updates user revision
        """

        profile_rev = self.profile.sso_rev

        self.profile.country = 'IT'
        self.profile.save()

        self.profile.refresh_from_db()

        self.assertEqual(self.profile.sso_rev, profile_rev + 1)

    def test_django_user_email_update_updates_profile_django_user_email(self):
        profile_rev = self.profile.sso_rev

        self.assertEqual(self.user.email, self.profile.django_user_email)

        self.user.email = 'new_email@example.com'
        self.user.save()

        self.assertEqual(self.profile.django_user_email, 'new_email@example.com')

        self.assertEqual(self.profile.sso_rev, profile_rev + 1)

    def test_django_user_username_update_updates_profile_django_user_username(self):
        profile_rev = self.profile.sso_rev

        self.assertEqual(self.user.username, self.profile.django_user_username)

        self.user.username = 'new_username'
        self.user.save()

        self.assertEqual(self.profile.django_user_username, 'new_username')

        self.assertEqual(self.profile.sso_rev, profile_rev + 1)



# app only accetta jwt, la valida via secret in setting file
# app persistence accetta jwt, valida e
"""
DJANGO_SSO_APP_SHAPES = ('app',
                         'app_persistence',
                         'app_apigateway',
                         'app_persistence_apigateway')
"""
import json
import uuid
import jwt
import responses

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.test import Client

from http.cookies import SimpleCookie

from rest_framework import status

from django_sso_app.backend.tests.factories import UserTestCase

from django_sso_app.core import app_settings

User = get_user_model()


class TestTokens(UserTestCase):
    def setUp(self):
        super(TestTokens, self).setUp()

        self.valid_jwt = jwt.encode(
            self.device0.get_jwt_payload(),
            'secret',
            app_settings.JWT_ALGORITHM
        ).decode('utf-8')

        self.invalid_jwt = jwt.encode(
            self.device0.get_jwt_payload(),
            'invalid_secret',
            app_settings.JWT_ALGORITHM
        ).decode('utf-8')

        self.valid_device_jwt = jwt.encode(
            self.device2.get_jwt_payload(),
            self.device2.apigw_jwt_secret,
            app_settings.JWT_ALGORITHM
        ).decode('utf-8')

    def test_app_shape_uses_settings_secret(self):
        client = Client()
        with self.settings(DJANGO_SSO_APP_SHAPE='app', DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            client.cookies = SimpleCookie({app_settings.JWT_COOKIE_NAME: self.valid_jwt})

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(self.profile.sso_id,)),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), str(self.profile.sso_id))

        client = Client()
        with self.settings(DJANGO_SSO_APP_SHAPE='app', DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            client.cookies = SimpleCookie({app_settings.JWT_COOKIE_NAME: self.invalid_jwt})

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(self.profile.sso_id,)),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN, 'invalid_jwt is valid')

    def test_app_apigateway_shape_uses_device_secret(self):
        with self.settings(DJANGO_SSO_APP_SHAPE='app_apigateway', DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            client = Client()
            client.cookies = SimpleCookie({app_settings.JWT_COOKIE_NAME: self.valid_device_jwt})

            response = client.get(
                reverse('django_sso_app_profile:rest-detail', args=(self.profile.sso_id,)),
                content_type='application/json',
                HTTP_X_CONSUMER_USERNAME=self.profile.sso_id
            )
            print('test_app_apigateway_shape_uses_device_secret RESP', response.data)
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), str(self.profile.sso_id))



import logging

from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import ObjectDoesNotExist
# from django.utils.encoding import smart_str
from django.utils.http import urlencode
from django.http import HttpResponseRedirect
from django.contrib.auth import get_user_model

from allauth.account.auth_backends import AuthenticationBackend as allauth_AuthenticationBackend

from .apps.users.utils import fetch_remote_user, create_local_user_from_remote_backend, \
                              update_local_user_from_remote_backend, create_local_user_from_jwt, \
                              create_local_user_from_apigateway_headers

from .permissions import is_authenticated
from .apps.profiles.models import Profile
from .apps.groups.utils import update_profile_groups
from .exceptions import ServiceSubscriptionRequiredException
from .apps.api_gateway.functions import get_apigateway_sso_id, get_apigateway_profile_groups_from_header
# from .apps.profiles.utils import get_or_create_user_profile

from . import app_settings
# from .utils import set_session_key, get_session_key

logger = logging.getLogger('django_sso_app.core')
User = get_user_model()


class DjangoSsoAppLoginAuthenticationBackend(allauth_AuthenticationBackend):
    def authenticate(self, request, **credentials):
        user = super(DjangoSsoAppLoginAuthenticationBackend, self).authenticate(request, **credentials)

        if is_authenticated(user):
            logger.debug('"{}" is authenticated'.format(user))

            return user


class DjangoSsoAppBaseAuthenticationBackend(ModelBackend):
    """
    See django.contrib.auth.backends.RemoteUserBackend.

    The jwt contains an expected payload such as the following

        {
          "sso_id": "<uuid>",
          "sso_rev": 0,
          "id": 1,
          "fp": "b28493a6a29a5a38973a8a3e3abe7f34",
          "iss": "nxhnkMPKcpRWaTKwNOvGcLcs5MHJINOg"
        }
    """

    def authenticate(self, request, **kwargs):
        """
        Returns user instance or None

        :param request:
        :param kwargs: django chooses the right authentication backend by parsing "authenticate" method signature
        :return:
        """
        raise NotImplementedError('authenticate')

    def try_update_user(self, sso_id, user, user_profile, encoded_jwt, decoded_jwt, **kwargs):
        raise NotImplementedError('try_update_user')

    def try_update_profile_groups(self, request, user):
        raise NotImplementedError('try_update_profile_groups')

    def redirect_to_service_subscription(self, user):
        _qs = urlencode({'next': app_settings.SERVICE_URL})
        _url = '{}{}?{}'.format(app_settings.BACKEND_URL, app_settings.LOGIN_URL, _qs)
        logger.info('User {} must agree to the Terms of Service,'
                    ' redirecting to {} ...'
                    .format(user, _url))
        response = HttpResponseRedirect(redirect_to=_url)

        raise ServiceSubscriptionRequiredException(response)

    def try_update_profile_subscriptions(self, user):
        # Redirect user to Term Of Service.
        # is_to_subscribe = getattr(user, '__dssoa__is_to_subscribe', False)
        if user and app_settings.SERVICE_SUBSCRIPTION_REQUIRED:
            logger.debug('update subscriptions check for "{}":"{}"'.format(app_settings.SERVICE_URL,
                                                                           app_settings.SERVICE_SUBSCRIPTION_REQUIRED))

            if app_settings.BACKEND_ENABLED:
                user_service_subscription = user.sso_app_profile.subscriptions.filter(
                    service__service_url=app_settings.SERVICE_URL).first()

                if ((user_service_subscription is None) or (user_service_subscription.is_unsubscribed)):
                    self.redirect_to_service_subscription(user)

            else:
                if app_settings.REPLICATE_PROFILE:
                    remote_user = getattr(user, '__dssoa__remote_user', None)
                    if remote_user is not None:
                        remote_profile_subscriptions = remote_user['profile']['subscriptions']
                        logger.debug('remote_profile_subscriptions "{}"'.format(remote_profile_subscriptions))

                        active_subscriptions = []
                        for s in remote_profile_subscriptions:
                            if not s['is_unsubscribed']:
                                active_subscriptions.append(s['service_url'])

                        if app_settings.SERVICE_URL not in active_subscriptions:
                            self.redirect_to_service_subscription(user)


class DjangoSsoAppApiGatewayAuthenticationBackend(DjangoSsoAppBaseAuthenticationBackend):
    """
    Authenticates by request CONSUMER_USERNAME header
    """

    backend_path = 'django_sso_app.core.backends.DjangoSsoAppApiGatewayAuthenticationBackend'

    def authenticate(self, request, consumer_username, encoded_jwt, decoded_jwt):
        logger.info('backend authenticating by apigateway consumer {}'.format(consumer_username))

        if consumer_username is None:
            logger.debug('consuner_usename not set, skipping authentication')
            return

        if app_settings.BACKEND_ENABLED:
            logger.debug('backend enabled')
            try:
                sso_id = get_apigateway_sso_id(consumer_username)
                profile = Profile.objects.get(sso_id=sso_id)
                user = profile.user

            except ObjectDoesNotExist:
                logger.debug('user with apigateway consumer username "{}" do not exists'.format(consumer_username))
                return
            else:
                logger.debug('user with apigateway consumer username "{}" EXISTS'.format(consumer_username))
                # allauth logic
                setattr(user, 'backend', self.backend_path)

        else:
            logger.debug('app enabled')
            try:
                sso_id = get_apigateway_sso_id(consumer_username)
                profile = Profile.objects.get(sso_id=sso_id)
                user = profile.user

            except ObjectDoesNotExist:
                logger.info('No profile with id "{}"'.format(sso_id))
                try:
                    user = self.try_replicate_user(request, sso_id, encoded_jwt)

                except Exception as e:
                    logger.warning('cannot replicate user because of "{}"'.format(e))
                    return

            else:
                if app_settings.REPLICATE_PROFILE:
                    if decoded_jwt is None:
                        logger.warning('decoded_jwt not set')
                        return

                    user = self.try_update_user(sso_id, user, profile, encoded_jwt, decoded_jwt, consumer_username)

                else:
                    # just updates user groups
                    logger.debug('Do not update profile')

            # updating relations
            if app_settings.MANAGE_USER_GROUPS:
                self.try_update_profile_groups(request, user)

            self.try_update_profile_subscriptions(user)

        return user

    def try_replicate_user(self, request, sso_id, encoded_jwt):
        logger.debug('try_replicate_user')

        if app_settings.REPLICATE_PROFILE:
            logger.info('Replicate user with sso_id "{}" from remote backend'.format(sso_id))
            # create local profile from SSO
            backend_user = fetch_remote_user(sso_id=sso_id, encoded_jwt=encoded_jwt)

            user = create_local_user_from_remote_backend(backend_user)

        else:
            # create local profile from headers
            logger.info('Do not replicate user, creating from headers')
            try:
                user = create_local_user_from_apigateway_headers(request)

            except Exception as e:
                logger.warning('cannot create user from headers because of "{}"'.format(e))
                return

        return user

    def try_update_user(self, sso_id, user, user_profile, encoded_jwt, decoded_jwt, consumer_username):
        logger.debug('try_update_user')

        rev_changed = user_profile.sso_rev < decoded_jwt['sso_rev']
        first_access = not user.is_active and not user_profile.is_unsubscribed

        if rev_changed or first_access:
            if rev_changed:
                logger.info('Rev changed from "{}" to "{}" for apigateway consumer "{}", updating ...'
                            .format(user_profile.sso_rev, decoded_jwt['sso_rev'],
                                    consumer_username))
            if first_access:
                logger.info('"{} apigateway consumer first access, updating ...'.format(consumer_username))

            # update local profile from SSO
            setattr(user, '__dssoa__apigateway_update', True)

            remote_user = fetch_remote_user(sso_id=sso_id, encoded_jwt=encoded_jwt)     
            user = update_local_user_from_remote_backend(remote_user=remote_user,
                                                         profile=user_profile)

            logger.info('{} updated with latest data from SSO'.format(user))

        else:
            logger.info('Nothing changed for apigateway consumer "{}"'.format(consumer_username))

        return user

    def try_update_profile_groups(self, request, user):
        logger.info('try_update_profile_groups for "{}"'.format(user))

        consumer_groups = request.META.get(
            app_settings.APIGATEWAY_CONSUMER_GROUPS_HEADER, None)
        profile_groups = get_apigateway_profile_groups_from_header(consumer_groups)

        update_profile_groups(user.sso_app_profile, profile_groups)


class DjangoSsoAppJwtAuthenticationBackend(DjangoSsoAppBaseAuthenticationBackend):
    """
    Authenticates by request jwt
    """

    backend_path = 'django_sso_app.core.backends.DjangoSsoAppApiGatewayAuthenticationBackend'

    def authenticate(self, request, encoded_jwt, decoded_jwt):
        logger.info('backend authenticating by request jwt')

        if encoded_jwt is None or decoded_jwt is None:
            logger.debug('request jwt not set, skipping authentication')
            return

        if app_settings.BACKEND_ENABLED:
            logger.debug('backend enabled')

            try:
                sso_id = decoded_jwt['sso_id']
                profile = Profile.objects.get(sso_id=sso_id)
                user = profile.user

            except ObjectDoesNotExist:
                logger.debug('user with sso_id "{}" does not exists'.format(sso_id))
                return

            else:
                # allauth logic
                setattr(user, 'backend', self.backend_path)
        else:
            logger.debug('app enabled')

            try:
                sso_id = decoded_jwt['sso_id']
                user_profile = Profile.objects.get(sso_id=sso_id)
                user = user_profile.user

                if app_settings.REPLICATE_PROFILE:
                    if decoded_jwt is None:
                        logger.warning('decoded_jwt not set')
                        return

                    logger.debug('try_update_user "{}" for apigateway consumer "{}"'.format(sso_id, sso_id))

                    user = self.try_update_user(sso_id, user, user_profile, encoded_jwt, decoded_jwt, sso_id)

                else:
                    # just updates user groups
                    logger.debug('Do not replicate profile')

            except ObjectDoesNotExist:
                try:
                    user = self.try_replicate_user(sso_id, encoded_jwt, decoded_jwt)

                except Exception as e:
                    logger.warning('cannot replicate user because of "{}"'.format(e))
                    return

            # update relations
            if app_settings.MANAGE_USER_GROUPS:
                self.try_update_profile_groups(request, user)

            self.try_update_profile_subscriptions(user)

        return user

    def try_replicate_user(self, sso_id, encoded_jwt, decoded_jwt):
        logger.debug('try_replicate_user')

        assert(encoded_jwt is not None and decoded_jwt is not None)

        if app_settings.REPLICATE_PROFILE:
            # create local profile from SSO
            logger.info('Replicate user with sso_id "{}" from remote backend'.format(sso_id))

            backend_user = fetch_remote_user(sso_id=sso_id,
                                             encoded_jwt=encoded_jwt)

            user = create_local_user_from_remote_backend(backend_user)

        else:
            # create local profile from jwt
            logger.info('Replicating user with sso_id "{}" from JWT'.format(sso_id))

            user = create_local_user_from_jwt(decoded_jwt)

            backend_user = decoded_jwt

        # setattr(user, 'sso_app_profile', get_or_create_user_profile(user, Profile, backend_user))

        return user

    def try_update_user(self, sso_id, user, user_profile, encoded_jwt, decoded_jwt):
        logger.debug('try_update_user')

        rev_changed = user_profile.sso_rev < decoded_jwt['sso_rev']
        first_access = not user.is_active and not user_profile.is_unsubscribed

        if rev_changed or first_access:
            if rev_changed:
                logger.info('Rev changed from "{}" to "{}" for "{}", updating ...'
                            .format(user_profile.sso_rev, decoded_jwt['sso_rev'],
                                    sso_id))
            if first_access:
                logger.info('"{}" first access, creating ...'.format(sso_id))

            ## update local profile from SSO

            logger.info('"{}" updated with latest data from SSO'.format(sso_id))

        else:
            logger.info('Nothing changed for "{}"'.format(sso_id))

        return user

    def try_update_profile_groups(self, request, user):
        # Update user groups
        if app_settings.REPLICATE_PROFILE:
            logger.info('try_update_profile_groups for "{}"'.format(user))

            remote_user_object = getattr(user, '__dssoa__remote_user')
            remote_user_object_groups = remote_user_object.get('groups', [])
            update_profile_groups(user.sso_app_profile, remote_user_object_groups)



import json
import responses
import jwt
from http.cookies import SimpleCookie
from django.contrib.auth import get_user_model
from django.test import Client
from django.urls import reverse
from django.contrib.auth.hashers import make_password
from django.test import TestCase, Client
from django.test.utils import override_settings

from rest_framework import status
from allauth.account.adapter import get_adapter
from rest_framework.authtoken.models import Token

from django_sso_app.core import app_settings

from django_sso_app.core.apps.profiles.models import Profile
from django_sso_app.core.apps.emails.models import EmailAddress

User = get_user_model()


class TestStatus(TestCase):
    def setUp(self):
        self.client = Client()
        adapter = get_adapter()

        self.email = "pippo@disney.com"
        self.username = self.username = "pippo"
        self.password = self.password = "paperina"
        self.hashed_password = make_password(self.password)

        self.valid_login = {
            'login': self.username,
            'password': self.password,
            'fingerprint': '123456'
        }

        self.new_user = User.objects.create(username=self.username, email=self.email, password=self.hashed_password)
        self.new_user_email = EmailAddress.objects.create(user=self.new_user,
                                                          email=self.email,
                                                          primary=True,
                                                          verified=True)

        self.device0 = adapter.add_user_profile_device(self.new_user, '000000')

        self.valid_jwt = jwt.encode(
            self.device0.get_jwt_payload(),
            'secret',
            app_settings.JWT_ALGORITHM
        ).decode('utf-8')

        staff_user_email = 'staff@example.com'

        self.staff_user_username = 'staff'
        self.staff_user_password = 'abc123456'
        self.staff_user = User.objects.create_user(username=self.staff_user_username, email=staff_user_email,
                                                   password=self.staff_user_password, is_staff=True)
        self.staff_user_email = EmailAddress.objects.create(user=self.staff_user,
                                                            email=self.staff_user.email,
                                                            primary=True,
                                                            verified=True)
        self.staff_user_token = Token.objects.create(user=self.staff_user)

        self.staff_user_valid_token_headers = {
            'content_type': 'application/json',
            'HTTP_AUTHORIZATION': 'Token {}'.format(self.staff_user_token.key)
        }

        self.staff_user_valid_login = {
            'login': self.staff_user.username,
            'password': self.staff_user_password,
            'fingerprint': '123456'
        }

        np_uuid = 'new-profile-uuid'
        print('new profile uuid', np_uuid)
        self.valid_new_profile = {
            "url": "http://localhost:8000/api/v1/auth/users/{}/".format(np_uuid),
            "sso_id": np_uuid,
            "sso_rev": 5,
            "date_joined": "2018-12-31T00:00:00+0000",
            "is_active": True,
            "email_verified": True,
            "groups": ['users'],
            "profile": {
                "url": "http://localhost:8000/api/v1/auth/profiles/{}/".format(np_uuid),
                "created_at": "2019-12-05T19:56:41+0000",
                "sso_id": np_uuid,
                "sso_rev": 5,
                "subscriptions": [
                    {
                        "service_url": "http://example.com",
                        "is_unsubscribed": False
                    }
                ],
                "is_unsubscribed": False,
                "username": "new_profile",
                "email": "paiuolo@gmail.com",
                "user": "http://localhost:8000/api/v1/auth/users/{}/".format(np_uuid),
                "first_name": "new",
                "last_name": "profile",
                "description": None,
                "picture": None,
                "birthdate": None,
                "latitude": None,
                "longitude": None,
                "country": "IT",
                "address": None,
                "language": "it"
            },
            "username": "new_profile",
            "email": "new_profile@example.com",
            "password": "pwd"
        }

        self.valid_new_profile_jwt_payload = {
                'id':999,
                'fingerprint': 'xxx',
                'sso_id': np_uuid,
                'sso_rev': 5
            }

        self.valid_new_profile_jwt = jwt.encode(
            self.valid_new_profile_jwt_payload,
            'secret',
            app_settings.JWT_ALGORITHM
        ).decode('utf-8')


    @responses.activate
    def test_switch_between_statuses(self):
        with self.settings(DJANGO_SSO_APP_SHAPE='backend_only'):

            client = Client()
            response = client.post(
                reverse('rest_login'),
                data=json.dumps(self.valid_login),
                content_type='application/json'
            )

            user_user_object = response.data.get('user')
            setattr(user_user_object, 'password', self.hashed_password)
            user_user_object = json.dumps(user_user_object)

            self.assertEqual(response.status_code, status.HTTP_200_OK, 'new user can not login')

        with self.settings(DJANGO_SSO_APP_SHAPE='backend_app'):
            client = Client()
            response = client.post(
                reverse('rest_login'),
                data=json.dumps(self.valid_login),
                content_type='application/json'
            )

            self.assertEqual(response.status_code, status.HTTP_200_OK, 'new user can not login')

        with self.settings(DJANGO_SSO_APP_SHAPE='app',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):

            profile_url = reverse('django_sso_app_profile:rest-detail', args=(self.valid_new_profile['sso_id'],))
            print('profile_url', profile_url)

            client = Client()
            client.cookies = SimpleCookie({app_settings.JWT_COOKIE_NAME: self.valid_new_profile_jwt})

            response = client.get(
                profile_url,
                content_type='application/json'
            )
            print('response', response.data)

            self.assertEqual(response.status_code, status.HTTP_200_OK, 'profile not created')

            self.assertEqual(Profile.objects.filter(sso_id=self.valid_new_profile['sso_id']).count(), 1,
                             'no profile model')

            self.assertEqual(response.data.get('sso_id'), str(self.valid_new_profile['sso_id']), 'sso_id differs')

        with self.settings(DJANGO_SSO_APP_SHAPE='app_persistence',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            # deleting pre generated user
            User.objects.get(sso_app_profile__sso_id=self.valid_new_profile['sso_id']).delete()

            profile_url = reverse('django_sso_app_profile:rest-detail', args=(self.valid_new_profile['sso_id'],))
            print('profile_url', profile_url)

            client = Client()

            mocked_url = app_settings.REMOTE_USER_URL.format(
                sso_id=self.valid_new_profile['sso_id'])  # + '?with_password=true'
            print('mocked_url', mocked_url)
            responses.add(responses.GET, mocked_url,
                          json=self.valid_new_profile, status=200)

            client.cookies = SimpleCookie({app_settings.JWT_COOKIE_NAME: self.valid_new_profile_jwt})

            response = client.get(
                profile_url,
                content_type='application/json'
            )

            self.assertEqual(Profile.objects.filter(sso_id=self.valid_new_profile['sso_id']).count(), 1,
                             'no profile model')

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), str(self.valid_new_profile['sso_id']), 'sso_id differs')

            created_profile = Profile.objects.get(sso_id=self.valid_new_profile['sso_id'])
            self.assertEqual(created_profile.groups.count(), 1, 'group not created')
            self.assertEqual(created_profile.groups.filter(name='users').count(), 1, 'group not same')

        with self.settings(DJANGO_SSO_APP_SHAPE='app_apigateway',
                           DJANGO_SSO_APP_SERVICE_URL='http://example.com'):
            # deleting pre generated user
            User.objects.get(sso_app_profile__sso_id=self.valid_new_profile['sso_id']).delete()

            profile_url = reverse('django_sso_app_profile:rest-detail', args=(self.valid_new_profile['sso_id'],))
            print('profile_url', profile_url)

            client = Client()

            client.cookies = SimpleCookie({app_settings.JWT_COOKIE_NAME: self.valid_new_profile_jwt})

            response = client.get(
                profile_url,
                content_type='application/json',
                HTTP_X_CONSUMER_USERNAME=self.valid_new_profile['sso_id'].replace('-', '_'),
                HTTP_X_CONSUMER_GROUPS='users'
            )
            print('response', response.data)
            self.assertEqual(Profile.objects.filter(sso_id=self.valid_new_profile['sso_id']).count(), 1,
                             'no profile model')

            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.data.get('sso_id'), str(self.valid_new_profile['sso_id']), 'sso_id differs')

            created_profile = Profile.objects.get(sso_id=self.valid_new_profile['sso_id'])
            self.assertEqual(created_profile.groups.count(), 1, 'group not created')
            self.assertEqual(created_profile.groups.filter(name='users').count(), 1, 'group not same')



import logging

from django.contrib.auth.signals import user_logged_in
from django.dispatch import receiver

logger = logging.getLogger('django_sso_app.core.apps.users.signals')


@receiver(user_logged_in)
def post_user_login(**kwargs):
    user = kwargs['user']

    logger.debug('Users, "{}" LOGGED IN!!!'.format(user))



import logging
# import inspect

from jwt.exceptions import InvalidSignatureError

from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured
from django.utils.deprecation import \
    MiddlewareMixin  # https://stackoverflow.com/questions/42232606/django
# -exception-middleware-typeerror-object-takes-no-parameters
from django.utils.encoding import smart_str
from django.conf import settings

from ..tokens.utils import get_request_jwt, jwt_decode
from ..permissions import is_authenticated, is_django_staff
from ..exceptions import ServiceSubscriptionRequiredException, RequestHasValidJwtWithNoDeviceAssociated
from ..utils import check_user_can_login  # , set_session_key
from ..apps.api_gateway.functions import get_apigateway_sso_id
from ... import app_settings

logger = logging.getLogger('django_sso_app.core.middleware')

ADMIN_URL = '/{}'.format(settings.ADMIN_URL)
SSO_ID_JWT_KEY = 'sso_id'


class DjangoSsoAppAuthenticationMiddleware(MiddlewareMixin):
    """
    See django.contrib.auth.middleware.RemoteUserMiddleware.
    """

    # Name of request header to grab username from.  This will be the key as
    # used in the request.META dictionary, i.e. the normalization of headers to
    # all uppercase and the addition of "HTTP_" prefix apply.
    header = app_settings.APIGATEWAY_CONSUMER_USERNAME_HEADER
    # Username for anonymous user
    anonymous_username = app_settings.APIGATEWAY_ANONYMOUS_CONSUMER_USERNAME

    # request info
    request_method = None
    request_path = None
    request_id = None
    request_ip = None

    @staticmethod
    def _remove_invalid_user(request):
        """
        Removes the current authenticated user in the request which is invalid.
        """
        logger.info('removing invalid user "{}"'.format(request.user))
        auth.logout(request)

    def process_request(self, request):

        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "SSO middleware requires the authentication middleware to be"
                " installed.  Edit your MIDDLEWARE setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the SsoMiddleware class.")


        # saving request info
        self.request_method = request.method
        self.request_path = request.path
        self.request_id = id(request)
        self.request_ip = request.META.get('REMOTE_ADDR', None)

        logger.info('--- "{}" request "{}" path "{}" method "{}"'.format(self.request_ip,
                                                                         self.request_id,
                                                                         self.request_path,
                                                                         self.request_method))

        # requesting user
        request_user = request.user


        # skipping django staff users
        if is_django_staff(request_user):
            logger.info('Skipping django staff user "{}"'.format(request_user))
            #  >= 1.10 has is_authenticated as parameter
            # If a staff user is already authenticated, we don't need to
            # continue
            # set_session_key(request, '__dssoa__user', request_user)
            return


        if self.request_path.startswith(ADMIN_URL):
            #  >= 1.10 has is_authenticated as parameter
            # If a staff user is already authenticated, we don't need to
            # continue
            if is_authenticated(request_user):
                _msg = 'Non staff user "{}" tried to enter admin path'.format(request_user)
                logger.warning(_msg)

                self._remove_invalid_user(request)
            else:
                logger.info('Skipping admin path')

            return

        """
        if self.request_path.startswith('/api/v1/passepartout'):
            logger.info('is passepartout path')
            # set_session_key(request, '__dssoa__is_passepartout_path', True)
        """

        consumer_username = None
        request_jwt = None
        verify_jwt = True

        if app_settings.APIGATEWAY_ENABLED:
            # getting sso_id by apigateway consumer username
            try:
                consumer_username = request.META[self.header]

                if consumer_username == self.anonymous_username:
                    if is_authenticated(request_user):
                        logger.info('consumer is anonymous, returning')
                        self._remove_invalid_user(request)
                    return

            except KeyError as e:
                # If specified header or jwt doesn't exist then remove any existing
                # authenticated user, or return (leaving request.user set to
                # AnonymousUser by the AuthenticationMiddleware).
                if is_authenticated(request_user):
                    logger.info('authentication middleware KeyError "{}"'.format(e))
                    self._remove_invalid_user(request)
                return

            else:
                sso_id = get_apigateway_sso_id(consumer_username)


        # validating request jwt

        try:
            request_jwt = get_request_jwt(request, encoded=False)

            if request_jwt is None:
                # If request has no JWT authentication skips
                logger.info('No JWT in request, skipping')
                return
            else:
                # decoding request JWT
                request_device, decoded_jwt = jwt_decode(request_jwt, verify=verify_jwt)

        except KeyError:
            logger.warning('Error decoding JWT "{}"'.format(request_jwt))

            if is_authenticated(request_user):
                self._remove_invalid_user(request)
            return

        except InvalidSignatureError:
            logger.warning('Invalid JWT signature "{}"'.format(request_jwt))

            if is_authenticated(request_user):
                self._remove_invalid_user(request)
            return

        else:
            if decoded_jwt is None:
                logger.warning('No decoded JWT for "{}"'.format(request_jwt))
            else:
                sso_id = decoded_jwt[SSO_ID_JWT_KEY]


        # If the user is already authenticated and that user is active and the
        # one we are getting passed in the headers, then the correct user is
        # already persisted in the session and we don't need to continue.

        # double check sso_id equality between received jwt and sso_app_profile
        if is_authenticated(request_user):
            # check declared sso_id
            if smart_str(request_user.sso_id) == sso_id:
                logger.debug('user is ok')
                #set_session_key(request, '__dssoa__user', request_user)
                return
            else:
                # An authenticated user is associated with the request, but
                # it does not match the authorized user in the header.
                logger.warning('sso_id differs! "{}" "{}"'.format(sso_id, request_user.sso_id))
                self._remove_invalid_user(request)
                return


        # We are seeing this user for the first time in this session, attempt
        # to authenticate the user.
        try:
            if app_settings.APIGATEWAY_ENABLED:
                user = auth.authenticate(request=request,
                                         consumer_username=consumer_username,
                                         encoded_jwt=request_jwt,
                                         decoded_jwt=decoded_jwt)
            else:
                user = auth.authenticate(request=request,
                                         encoded_jwt=request_jwt,
                                         decoded_jwt=decoded_jwt)

        except ServiceSubscriptionRequiredException as e:
            logger.info('ServiceSubscriptionRequiredException')
            if is_authenticated(request_user) and app_settings.APP_ENABLED:
                self._remove_invalid_user(request)

            return e.response

        except RequestHasValidJwtWithNoDeviceAssociated:
            logger.exception('RequestHasValidJwtWithNoDeviceAssociated')

            if is_authenticated(request_user):
                self._remove_invalid_user(request)

            return

        except Exception as e:
            logger.exception('generic middleware exception "{}"'.format(e))
            if is_authenticated(request_user):
                self._remove_invalid_user(request)
            return

        # finishing

        if user is not None:
            # User is valid.  Set request.user and persist user in the session
            # by logging the user in.

            try:
                check_user_can_login(user)

            except Exception as e:
                logger.exception('Authentication middleware check_user_can_login exception "{}"'.format(e))

                if is_authenticated(request_user):
                    self._remove_invalid_user(request)

                return

            else:
                logger.info('User "{}" authenticated successfully!'.format(user))

                # set request user
                setattr(request, 'user', user)  # !!

    def process_response(self, request, response):
        logger.info('--- "{}" response "{}" path "{}" method "{}" ({})'.format(self.request_ip,
                                                                               self.request_id,
                                                                               self.request_path,
                                                                               self.request_method,
                                                                               response.status_code))

        return response



import logging

from django.contrib.auth import get_user_model
from django.db.models.signals import m2m_changed
from django.db.models.signals import pre_delete, pre_save
from django.dispatch import receiver

from allauth.account.adapter import get_adapter

from ....permissions import is_django_staff
from ...devices.models import Device
from ...groups.models import Group
from ...profiles.models import Profile
from ..utils import delete_apigw_consumer, create_apigw_consumer_jwt, delete_apigw_consumer_jwt, \
                    create_apigw_consumer_acl, delete_apigw_consumer_acl, get_profile_apigw_consumer_id

logger = logging.getLogger('django_sso_app.core.apps.api_gateway')

User = get_user_model()
adapter = get_adapter()

# device

@receiver(pre_save, sender=Device)
def create_device_apigw_jwt(sender, instance, **kwargs):
    profile = instance.profile
    user = profile.user

    if not is_django_staff(user):
        apigateway_consumer_id = get_profile_apigw_consumer_id(profile)

        logger.info('Profile "{}" has apigateway_consumer_id "{}"'.format(profile, apigateway_consumer_id))

        if instance._state.adding:
            device = instance

            logger.info('Creating apigw JWT for Device "{}"'.format(device))

            status_code, r1 = create_apigw_consumer_jwt(device.profile)
            if status_code != 201:
                # delete_apigw_consumer(username)
                logger.error(
                    'Error ({}) creating apigw consumer JWT, "{}"'.format(
                        status_code, r1))
                raise Exception(
                    'Error ({}) creating apigw consumer jwt'.format(status_code))

            device.apigw_jwt_id = r1.get('id')
            device.apigw_jwt_key = r1.get('key')
            device.apigw_jwt_secret = r1.get('secret')


@receiver(pre_delete, sender=Device)
def delete_device_api_jwt(sender, instance, **kwargs):
    device = instance
    logger.info('Deleting apigw JWT for Device "{}", jwt_id: "{}"'.format(device, device.apigw_jwt_id))

    status_code, r1 = delete_apigw_consumer_jwt(device.profile, device.apigw_jwt_id)
    logger.info('Kong JWT deleted ({0}), {1}'.format(status_code, r1))
    if status_code >= 300:
        if status_code != 404:
            # delete_apigw_consumer(username)
            logger.error('Error ({}) Deleting apigw JWT for Device "{}", "{}"'.format(status_code, device, r1))

            raise Exception(
                "Error deleting apigw consumer jwt, {}".format(status_code))

# profile

@receiver(pre_delete, sender=Profile)
def delete_apigw_user_consumer(sender, instance, **kwargs):
    user = instance.user

    if not is_django_staff(user):
        profile = instance
        logger.info('Profile "{}" will be deleted, removing api gateway consumer'.format(profile))

        status_code, r1 = delete_apigw_consumer(profile)
        logger.info('Api gateway consumer deletion: ({0}), {1}'.format(status_code, r1))

        if status_code >= 300:
            if status_code != 404:
                # delete_apigw_consumer(username)
                logger.error(
                    'Error ({0}) Deleting apigw consumer for User {1}, {2}'.format(
                        status_code, profile, r1))
                raise Exception(
                    "Error deleting apigw consumer, {}".format(status_code))

# group

@receiver(m2m_changed)
def signal_handler_when_profile_is_added_to_or_removed_from_group(action, instance, pk_set, model, **kwargs):
    is_loaddata = getattr(instance, '__dssoa__loaddata', False)

    if model == Group and instance.__class__ == Profile:
        profile = instance
        user = profile.user

        if not is_django_staff(user):
            if action == 'pre_add':
                for pk in pk_set:
                    group = Group.objects.get(id=pk)
                    logger.info('creating api gateway acl group "{}" for profile "{}"'.format(group, profile))

                    status_code, resp = create_apigw_consumer_acl(profile, group.name)

                    if status_code == 201:
                        logger.info('api gateway acl goup created {}, {}'.format(status_code, resp))

                        if not is_loaddata:
                            profile.update_rev(True)
                    else:
                        logger.exception('error ({}) creating api gateway acl group'.format(status_code))

            elif action == 'pre_remove':
                for pk in pk_set:
                    group = Group.objects.get(id=pk)
                    logger.info('deleting apigw acl group "{}" for "{}"'.format(group, profile))

                    status_code, resp = delete_apigw_consumer_acl(profile, group.name)

                    if status_code == 200:
                        logger.info('"{}" DELETED, updating profile sso_rev'.format(group))

                        if not is_loaddata:
                            profile.update_rev(True)
                    else:
                        logger.exception('error ({}) creating api gateway acl group'.format(status_code))



import logging

from django.dispatch import receiver
from django.db.models.signals import post_save, pre_save

from ..models import EmailAddress

logger = logging.getLogger('django_sso_app.core.apps.emails.signals')


@receiver(post_save, sender=EmailAddress)
def emailaddress_updated(sender, instance, created, **kwargs):
    if kwargs['raw']:
        # https://github.com/django/django/commit/18a2fb19074ce6789639b62710c279a711dabf97
        return

    if instance.primary and instance.verified:
        # deleting all other user email addresses
        logger.info('Email "{}" primary and verified, deleting other user emails'.format(instance))
        user = instance.user
        user.emailaddress_set.exclude(pk=instance.pk).all().delete()



# profile filed update updates rev

from django.contrib.auth import get_user_model

from django_sso_app.backend.tests.factories import UserTestCase

User = get_user_model()


class UserModelTestCase(UserTestCase):

    def test_django_superuser_user_creation_creates_sso_app_profile_with_sso_id_equals_to_username(self):
        self.assertEqual(self.admin_user.sso_id, self.admin_user.username)

    def test_django_staff_user_creation_creates_sso_app_profile_with_sso_id_equals_to_username(self):
        self.assertEqual(self.staff_user.sso_id, self.staff_user.username)



from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password
from django.test import TestCase, Client

from allauth.account.adapter import get_adapter

from rest_framework.authtoken.models import Token

from django_sso_app.core.apps.emails.models import EmailAddress
from django_sso_app.core.apps.services.models import Service, Subscription

User = get_user_model()


class UserTestCase(TestCase):
    def setUp(self):
        self.client = Client()
        self.adapter = get_adapter()

        email = self.email = "pippo@disney.com"
        username = self.username = "pippo"
        password = self.password = "paperina"
        updated_password = "pippapippa"

        username2 = "cenerentola"
        email2 = "cenerentola@disney.com"
        password2 = "renecentola"

        self.admin_user = User.objects.create_user(username='admin', email='admin@example.com', password='admin',
                                                   is_superuser=True, is_staff=True)

        self.service = Service.objects.create(service_url='http://example.com', name='example.com')
        self.service2 = Service.objects.create(service_url='http://disney.org', name='disney.org',
                                               subscription_required=True)

        self.user = User.objects.create_user(username=username, email=email, password=password)
        self.user_email = EmailAddress.objects.create(user=self.user,
                                                      email=email,
                                                      primary=True,
                                                      verified=True)
        self.profile = self.user.sso_app_profile

        self.user_subscription = Subscription.objects.create(profile=self.profile,
                                                             service=self.service)




        self.user2 = User.objects.create_user(username=username2, email=email2, password=password2)

        self.user2_email = EmailAddress.objects.create(user=self.user2,
                                                       email=email2,
                                                       primary=True,
                                                       verified=True)
        self.profile2 = self.user2.sso_app_profile


        staff_email = staff_user_email = 'staff@example.com'

        self.staff_user_username = 'staff'
        self.staff_user_password = 'abc123456'
        self.staff_user = User.objects.create_user(username=self.staff_user_username, email=staff_user_email,
                                                   password=self.staff_user_password, is_staff=True)
        self.staff_user_email = EmailAddress.objects.create(user=self.staff_user,
                                                            email=self.staff_user.email,
                                                            primary=True,
                                                            verified=True)
        self.staff_user_token = Token.objects.create(user=self.staff_user)

        self.staff_user_valid_token_headers = {
            'content_type': 'application/json',
            'HTTP_AUTHORIZATION': 'Token {}'.format(self.staff_user_token.key)
        }

        self.valid_new_external_user = {
            'email': 'cane@gmail.com',
            'username': 'cane',
            'password': 'pippopluto',
            'profile': {
                'first_name': 'Pluto',
                'last_name': 'Cagnaccio',
                'ssn': '123456',
                'phone': '+39144144144',
                'description': 'Bau bau..',
                'picture': None,
                'birthdate': '1940-01-01',
                'latitide': 11.12,
                'longitude': 30.1,
                'country': 'it',
                'address': 'Viale dei cani 11, 11040, Tripoli',
                'language': 'it',
                'subscriptions': [
                    {
                        'service': {
                            'name': self.service2.name
                        }
                    }
                ]
            }
        }

        self.valid_profile_update = self.valid_new_external_user['profile']

        self.valid_new_profile = {
            'email': 'paperino@gmail.com',
            'username': 'paperino',
            'password1': 'supercacca',
            'password2': 'supercacca',
            'referrer': 'http://example.com'
        }

        self.valid_new_profile_case_sensitive = {
            'email': 'paperino2@gmail.com',
            'username': 'Paperino',
            'password1': 'supercacca',
            'password2': 'supercacca',
            'referrer': 'http://example.com'
        }

        self.invalid_new_profile = {
            'password': 'nonnapapera',
            'referrer': 'http://example.com'
        }

        self.valid_new_profile_hashed_pw = {
            'email': 'paperino@gmail.com',
            'username': 'paperino',
            'password': make_password('nonnapapera'),
            'referrer': 'http://example.com'
        }

        self.valid_new_profile_login = {
            'login': 'paperino',
            'password': 'nonnapapera',
            'fingerprint': '123456'
        }

        self.staff_user_valid_login = {
            'login': self.staff_user.username,
            'password': self.staff_user_password,
            'fingerprint': '123456'
        }

        self.valid_login = {
            'login': username,
            'password': password,
            'fingerprint': '123456'
        }

        self.valid_staff_login = {
            'login': self.staff_user_username,
            'password': self.staff_user_password,
            'fingerprint': '123456'
        }

        self.valid_admin_login = {
            'login': 'admin',
            'password': 'admin'
        }

        self.valid_unsubscription = {
            'password': password
        }

        self.valid_email_login = {
            'login': email,
            'password': password,
            'fingerprint': '123456'
        }

        self.valid_password_update = {
            'password': updated_password
        }

        self.valid_email_update = {
            'email': 'pippo2@gmail.com'
        }

        self.invalid_unconfirmed_email_login = {
            'login': self.valid_email_update['email'],
            'password': password,
            'fingerprint': '123456'
        }

        self.valid_username_login = {
            'login': username,
            'password': password,
            'fingerprint': '123456'
        }

        self.invalid_login = {
            'login': 'topolino@disney.com',
            'password': ''
        }

        self.valid_login_after_password_change = {
            'login': username,
            'password': updated_password,
            'fingerprint': '123456'
        }

        self.device0 = self.adapter.add_user_profile_device(self.user, '000000')
        self.device1 = self.adapter.add_user_profile_device(self.user, '111111')
        self.device2 = self.adapter.add_user_profile_device(self.user, '222222')
        self.device2.apigw_jwt_secret='secret'
        self.device2.save()

        self.migrated_user_username = 'supercacca'
        self.migrated_user_email = 'supercacca@supercacca.com'
        self.migrated_user_password = 'supercacca'
        self.migrated_user_hashed_password = 'pbkdf2_sha256$15000$KCmb0T1BofB7$fV4zlS1mFUC8HYhrgnAtEfvFzB1pgnyDloidsDLYVag='

        self.valid_migrated_user_login = {
            'login': self.migrated_user_username,
            'password': self.migrated_user_password,
            'fingerprint': '123'
        }




import unittest

from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.chrome.options import Options

import time, sys

from django.test import LiveServerTestCase
from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from ...core.apps.services.models import Service


class FunctionalTest(StaticLiveServerTestCase):
    @classmethod
    def setUpClass(cls):
        for arg in sys.argv:
            if 'liveserver' in arg:
                cls.server_url = 'http://' + arg.split('=')[1]
                return
        super().setUpClass()
        cls.server_url = cls.live_server_url

    #def tearDownClass(cls):
    #    if cls.server_url == cls.live_server_url:
    #        super().tearDownClass()

    def setUp(self):
        #self.browser = webdriver.Chrome()
        
        options = webdriver.ChromeOptions()
        options.add_experimental_option('prefs', {'intl.accept_languages': 'en,en_US'})
        options.add_argument("start-maximized")
        options.add_argument("disable-infobars")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--no-sandbox")
        options.add_experimental_option("useAutomationExtension", False)
        self.browser = webdriver.Chrome(options=options)
        
        #self.browser = webdriver.FirefoxDriver()
        self.browser.implicitly_wait(1)
        
        self.service = Service.objects.create(name="example", url="https://www.example.com")
        
    def tearDown(self):
        self.browser.close()

    def identifica_titolo(self, titolo):
        # riconosce la pagina del partito dal titolo
        self.assertIn(titolo, self.browser.title)

    def tag_contiene_testo(self, tag, testo, in_testa=False):
        """
        in_testa ==> testo == lista[id_lista][0].text
        """
        elementi = self.browser.find_elements_by_tag_name(tag)
        self.assertTrue(len(elementi)>0, 'lista vuota')
        if in_testa:
            self.assertEqual(
                testo, elementi[0].text,
                'Elemento "{}" non in testa'.format(testo)
            )
        else:
            self.assertIn(
                testo, [el.text for el in elementi]
            )

    def compila_input_by_name(self, name_input, testo, enter=True, el='input'):
        input_x = self.browser.find_element_by_xpath('//'+el+'[@name="'+name_input+'"]')

        input_x.send_keys(testo)
        if enter:
          input_x.send_keys(Keys.ENTER)

        return input_x

    def compila_input_by_id(self, eid, testo, enter=True):
        input_x = self.browser.find_element_by_id(eid)
        input_x.send_keys(testo)
        if enter:
          input_x.send_keys(Keys.ENTER)
        return input_x

    def click_link(self, url):
        link = self.browser.find_element_by_xpath('//a[@href="'+url+'"]')
        print('\n LINKS elementi', link)
        link.click()
        
    def find_by_tag_class_and_text(self, tag, css, text):
        return self.browser.find_element_by_xpath("//{0}[contains(@class, '{1}') and text()='{2}']".format(tag, css, text))

    def find_by_tag_and_name(self, tag, name):
        return self.browser.find_element_by_xpath('//{0}[@name="{1}"]'.format(tag, name))

    def find_by_tag_and_class(self, tag, cls):
        return self.browser.find_element_by_xpath("//{0}[contains(@class, '{1}')]".format(tag, cls))
      
    def press_next_button(self, i):
        self.browser.find_elements_by_xpath('//{0}[@name="{1}"]'.format('button', 'next_step'))[i].click()


if __name__ == '__main__':
    try:
        unittest.main(warnings='ignore')

    except KeyboardInterrupt:
        print('Term signal received')
        exit(0)



import time
import os
import re
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from django.core import mail
from django.conf import settings

from .base import FunctionalTest
from ...core import app_settings


class NuovoVisitatoreTest(FunctionalTest):

    def test_puo_registrarsi_al_sito(self):
        step = 0
        # L'utente si connette all'indirizzo della pagina
        app_url = self.server_url + '?next=' + self.service.service_url
        self.browser.get(app_url)
        time.sleep(6)

        # riconosce la pagina
        self.identifica_titolo('Django SSO App')

        self.click_link('/login/')

        # che gli chiede di inserire le credenziali di accesso
        self.tag_contiene_testo('p', 'If you have not created an account yet, then please sign up first.')

        # prova a loggarsi
        self.compila_input_by_name('login', 'pippo@disney.com')

        self.compila_input_by_name('password', 'paperina')

        # vede che non è possibile senza creare un nuovo utente
        self.tag_contiene_testo('li', 'The e-mail address and/or password you specified are not correct.')

        # segue il link di registrazione
        self.click_link('/signup/')

        # inserisce email giusta
        self.compila_input_by_name('email', 'pippo@disney.com')

        # inserisce username
        self.compila_input_by_name('username', 'pippo')

        # inserisce password
        self.compila_input_by_name('password1', 'paperina')

        # conferma password
        self.compila_input_by_name('password2', 'paperina')

        # mette nome
        time.sleep(50)
        self.compila_input_by_name('first_name', 'Pippo')

        # mette cognome
        self.compila_input_by_name('last_name', 'Goofy')

        # mette description
        self.compila_input_by_name('description', 'Yuk! Yuk yuk yuk yuk!! YUK!', el="textarea")

        # inserisce immagine
        avatar_path = os.path.join(settings.ROOT_DIR, 'backend', 'tests', 'assets', 'avatar.png')
        self.browser.find_element_by_name('picture').send_keys(avatar_path)

        # inserisce posizione
        campo = self.compila_input_by_name('address', 'viale ettore andreis 74 desenzano', el="textarea")
        self.compila_input_by_name('latitude', '0')
        self.compila_input_by_name('longitude', '0')
        self.compila_input_by_name('country', 'it')

        # Lo si informa dell'avvenuta registrazione
        self.tag_contiene_testo('p', "We have sent an e-mail to you for verification. "
                                     "Follow the link provided to finalize the signup process. "
                                     "Please contact us if you do not receive it within a few minutes.")

        # Controlla la casella della posta in arrivo e segue il link di attivazione
        self.assertEqual(len(mail.outbox), 1)
        message = str(mail.outbox[0].message())
        print('MESSAGE!', message)
        activate_email_url = re.search("(?P<url>https?://.[^\s]+)", message).group("url")
        activate_email_url = activate_email_url.replace('http://' + app_settings.APP_DOMAIN, '')

        print('EMAIL_URL', activate_email_url, self.server_url + activate_email_url)
        self.browser.get(self.server_url + activate_email_url)

        time.sleep(1)

        # il link di attivazione lo redirige alla pagina di login
        self.tag_contiene_testo('p', 'If you have not created an account yet, then please sign up first.')

        # inserise le credenziali
        self.compila_input_by_name('login', 'pippo@disney.com')
        self.compila_input_by_name('password', 'paperina')

        time.sleep(2)

        self.tag_contiene_testo('h1', 'Welcome pippo')

        print('\n\nEOT')
        time.sleep(3)



import json

import jwt
from django.urls import reverse
from rest_framework import status

from ...users.tests.test_views import UserTestCase


class TestSubscriptions(UserTestCase):

    def test_user_can_list_subscriptions(self):
        """
        """
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        response2 = self.client.get(
            reverse('django_sso_app_service:rest-subscription-list'),
            content_type='application/json'
        )
        print('RESP!!!', response2.json())
        print('dfasdfdsf', self.profile.subscriptions.all())

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response2.data['results']), 1)
        self.assertEqual(response2.data['results'][0].get('service').get('service_url'), self.service.service_url)


    def test_user_can_retrieve_subscription(self):
        response = self.client.post(
            reverse('rest_login'),
            data = json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        response2 = self.client.get(
            reverse('django_sso_app_service:rest-subscription-detail', args=(self.user_subscription.id,)),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(response2.data.get('is_unsubscribed'), False)



class TestServices(UserTestCase):

    def test_user_can_list_services(self):
        """
        """
        response = self.client.post(
            reverse('rest_login'),
            data = json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        response2 = self.client.get(
            reverse('django_sso_app_service:rest-list'),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response2.data['results']), 2)

    def test_user_can_retrieve_service(self):
        """
        """
        response = self.client.post(
            reverse('rest_login'),
            data = json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        response2 = self.client.get(
            reverse('django_sso_app_service:rest-detail', args=(self.service.id, )),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(response2.data.get('service_url'), self.service.service_url)

    def test_user_can_not_subscribe_profile_to_service_twice(self):
        """
        """
        response = self.client.post(
            reverse('rest_login'),
            data = json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        response2 = self.client.post(
            reverse('django_sso_app_service:rest-subscription', args=(self.service.id, )),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_409_CONFLICT)
        self.assertEqual(response2.data, 'Already subscribed.')

    def test_user_can_subscribe_to_service_with_rev_incremented_and_devices_deleted(self):
        """
        """
        first_rev = self.profile.sso_rev
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies
        first_devices = self.profile.devices.all()
        first_devices = [x.id for x in first_devices]

        response2 = self.client.post(
            reverse('django_sso_app_service:rest-subscription', args=(self.service2.id, )),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_201_CREATED)

        self.profile.refresh_from_db()

        self.assertEqual(response2.data, 'Successfully subscribed.')

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev+1, 'rev not incremented')

        response3 = self.client.get(
            reverse('django_sso_app_service:rest-list'),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response3.data['results']), 2)

        cookies2 = response2.cookies
        jwt_cookie2 = cookies2.get('jwt').value
        unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

        print('\n\n\n Unv pay', unverified_payload2)

        self.assertEqual(unverified_payload2.get('sso_rev'), second_rev, 'cookie rev NOT changed after login')

        second_devices = self.profile.devices.all()
        second_devices = [x.id for x in second_devices]

        print('\n\n DEVICES', first_devices, second_devices)
        self.assertEqual(len(second_devices), 1)

    def test_user_can_unsubscribe_profile_from_service_with_rev_incremented(self):
        first_rev = self.profile.sso_rev

        response = self.client.post(
            reverse('rest_login'),
            data = json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        response2 = self.client.post(
            reverse('django_sso_app_service:rest-unsubscription', args=(self.service.id, )),
            data=json.dumps(self.valid_unsubscription),
            content_type='application/json'
        )
        cookies2 = response2.cookies

        print('\n\nRISPOSTA', response2.json())

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        self.assertEqual(response2.data, 'Successfully unsubscribed.')

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev+1, 'rev not incremented on service unsubscription')

        jwt_cookie = cookies2.get('jwt')
        self.assertNotEqual(jwt_cookie, None, 'no new jwt received')

        unverified_payload = jwt.decode(jwt_cookie.value, None, False)

        print('\n\n\n Unv pay', unverified_payload)

        self.assertEqual(unverified_payload.get('sso_rev'), second_rev, 'rev NOT updated in JWT')

    def test_staff_user_can_subscribe_profile_to_service_with_profile_rev_incremented(self):
        first_rev = self.profile.sso_rev

        response = self.client.post(
            reverse('django_sso_app_service:rest-subscription', args=(self.service2.id, )),
            data=json.dumps({'sso_id': str(self.profile.sso_id)}),
            **self.staff_user_valid_token_headers
        )
        cookies2 = response.cookies

        print('\n\nRISPOSTA', response.json())

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.data, 'Successfully subscribed.')

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev

        self.assertEqual(second_rev, first_rev+1, 'rev not incremented on service subscription')




import json

import jwt
from django.contrib.auth import get_user_model
from django.test import Client
from django.urls import reverse
from rest_framework import status

from django_sso_app.backend.tests.factories import UserTestCase

from .... import app_settings

User = get_user_model()


class TestProfileRetrieve(UserTestCase):

    def test_user_can_retrieve_profile_with_subscriptions(self):
        """
          Registration via rest_auth view
        """
        print('FIRST', json.dumps(self.valid_login))

        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = self.client.get(
            reverse('django_sso_app_profile:rest-detail', args=(self.profile.sso_id,)),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        print('RISPSTA2', response2.data)

        self.assertEqual(response2.data.get('subscriptions'),
                         [{'service_url': el.service.service_url, 'is_unsubscribed': el.is_unsubscribed} for el in
                          self.profile.subscriptions.all()])


class TestProfileUpdate(UserTestCase):

    def test_profile_update_refreshes_and_returns_current_device_jwt_and_deletes_other_devices(self):
        """
        Profile update refreshes caller jwt and deletes all but caller devices
        """
        first_rev = self.profile.sso_rev
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        response2 = self.client.patch(
            reverse('django_sso_app_profile:rest-detail', args=[self.profile.sso_id]),
            data=json.dumps(self.valid_profile_update),
            content_type='application/json'
        )
        cookies2 = response2.cookies

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev
        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertTrue(len(cookies2.keys()) > 0, 'no jwt cookie set')

        devices = self.profile.devices.all()

        self.assertEqual(devices.count(), 1, 'other devices not deleted')

        self.assertEqual(devices.first().fingerprint, self.valid_login['fingerprint'], 'device not updated')

        jwt_cookie = cookies.get('jwt').value
        jwt_cookie2 = cookies2.get('jwt').value

        # print('\nRISPOSTA, user rev', self.profile.sso_rev, '\n\nTOKEN', jwt_cookie)

        unverified_payload = jwt.decode(jwt_cookie, None, False)
        unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

        self.assertEqual(unverified_payload2.get('sso_rev'), unverified_payload.get('sso_rev') + 1,
                         'rev not incremented in cookie')

        profile_object = self.valid_profile_update
        for f in app_settings.PROFILE_FIELDS:
            model_val = getattr(self.user.sso_app_profile, f, None)
            object_val = profile_object.get(f, None)

            if f == 'birthdate':
                model_val = str(model_val)
            elif f == 'country':
                object_val = object_val.upper()
                model_val = str(model_val).upper()

            self.assertEqual(model_val, object_val)


class TestCompleteUnsubscription(UserTestCase):

    def test_user_can_completely_unsubscribe_deleting_devices_and_disables_login(self):
        """
        Profile update refreshes caller jwt and deletes all but caller devices
        """
        first_rev = self.profile.sso_rev
        response = self.client.post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        cookies = response.cookies

        url2 = reverse('django_sso_app_profile:rest-complete-unsubscription', args=[self.profile.sso_id])

        response2 = self.client.post(
            url2,
            data=json.dumps(self.valid_unsubscription),
            content_type='application/json'
        )
        cookies2 = response2.cookies

        self.assertEqual(response2.status_code, status.HTTP_200_OK)

        self.profile.refresh_from_db()
        second_rev = self.profile.sso_rev
        self.assertEqual(second_rev, first_rev + 1, 'rev not incremented')

        self.assertTrue(len(cookies2.keys()) > 0, 'no jwt cookie set')

        devices = self.profile.devices.all()

        self.assertEqual(devices.count(), 0, 'devices not deleted')

        self.assertEqual(response2.data, 'Successfully completely unsubscribed.')

        response3 = Client().post(
            reverse('rest_login'),
            data=json.dumps(self.valid_login),
            content_type='application/json'
        )
        print('RESP3!!!', response3.cookies.items(), self.valid_login, response3.data, response3.status_code)

        self.assertEqual(response3.status_code, status.HTTP_400_BAD_REQUEST)



from ..core.settings.common import *


DJANGO_SSO_APP_DJANGO_AUTHENTICATION_BACKENDS = [
    # requests
    'django_sso_app.core.backends.DjangoSsoAppApiGatewayAuthenticationBackend',
    'django_sso_app.core.backends.DjangoSsoAppJwtAuthenticationBackend'
]

DJANGO_SSO_APP_BACKEND_DOMAINS = env('DJANGO_SSO_APP_BACKEND_DOMAINS', default=[APP_DOMAIN])
